#!/usr/local/bin/perl -w
use strict;
use CGI qw(:standard);
use CGI::Carp qw(fatalsToBrowser);
use Time::Local;

my $listdata = 'listdata.txt';  # data file
my @data = ();                  # array of hashes

# global default settings
my $sortby = 'prior';           # order to sort list
my $showdone = 1;               # show done items?  (1 == yes)
 
&init();
&process();
 
sub init {
    # get the current date, put in in MM/DD/YY format
    my ($day,$month,$year) = 0;
    (undef,undef,undef,$day,$month,$year) = localtime(time);
    $month++;                     # months start from 0 
    $year += 1900;                # Perl years are years since 1900;
    # this keep us from getting bit by Y2K
    my $date = "$month/$day/$year";
    # open & read data file
    &read_data();
 
    # start HTML
    print header;
    print start_html('My To Do List');
    print "<h1 align=\"center\"><font face=\"Helvetica,Arial\">";

    print "To Do List</font></h1>\n";
    print "<h2 align=\"center\"><font face=\"Helvetica,Arial\">";
    print "$date</font></h2>\n";
    print "<hr />\n";
    print "<form method=\"post\">\n";
}
 
sub process {
    my $dateerror = 0;                # error in date format in old list
    my $newerror = 0;                # error in date format in new item
 
    # main switching point.  There are 2 choices:
    # no parameters, for displaying defaults
    # any parameters: update, add item if necessary, write and display
    if (!param()) {                # first time only
        &display_all();
    } else {                        # handle buttons
        &remove_selected();
        $dateerror = &update_data(); # update existing changes, if any
 
        # add items
        if (defined param('additems')) {
            $newerror = &check_date(param('newdate'));
            if (!$newerror) {
                &add_item();
            }
        }
 
        &write_data();
        &display_all($dateerror,$newerror);
    }
 
    print end_html;
}
 
# read data file into array of hashes
sub read_data {
    open(DATA, $listdata) or die "Can't open data file: $!";
    my %rec = ();
    while (<DATA>) {
        chomp;
        if ($_ =~ /^\#/) {
            next;
        }
        if ($_ ne '---' and $_ ne '') { # build the record
            my ($key, $val) = split(/=/,$_,2);
            $rec{$key} = $val;
        } else {                # end of record
            push @data, { %rec };
            %rec = ();
        }
    }
    close(DATA);
}

sub display_all {
    my $olderror = shift;        # has an error occurred?
    my $newerror = shift;
 
    if ($olderror or $newerror) {
        print "<p><font color=\"red\"><b>Error:  Dates marked with *** ";
        print "not in right format (use MM/DD/YYYY)</b></font></p>\n";
    }

    print "<table width=\"75%\" align=\"center\">\n";
    print "<tr bgcolor=\"silver\"><th>Done?</th><th>Priority</th>";
    print "<th>Date Due</th><th align=\"left\">Description</th>";
    print "<th>Remove?</th></tr>\n";

    # determine sort type (numeric or string) based on $sortby
    my @sdata = ();

    # sort the array of hashes based on value of $sortby
    if ($sortby eq 'date') {        # special date sort
        @sdata = sort {&date2time($a->{'date'}) <=> 
                       &date2time($b->{'date'})} @data;
    } else {                        # regular text/priority sort
        @sdata = sort {$a->{$sortby} cmp $b->{$sortby}} @data;
    }

    # print each item in order
    foreach (@sdata) {
        &display_data(%$_);        # pass in record
    }

    print "</table>\n";

    # preference table, with state preserved
    print "<p><table width=\"75%\" align=\"center\">\n";
    print "<tr><td align=\"center\"><b>Sort By:</b>";
    print "<select name=\"sortby\">\n";

    my @sort_options = ('prior', 'date', 'desc');
    my @sort_option_names = ('Priority', 'Date', 'Description');

    for (my $i = 0; $i < @sort_options; $i++)
    {
        # get current val of sortby, show menu
        print "<option value=\"$sort_options[$i]\" ";
        if ($sortby eq $sort_options[$i]) {
            print "selected>";
        } else {
            print ">";
        }
        print "$sort_option_names[$i]</option>\n";
    }

    print "</select></td>\n";
   
    # get current val of showdone, show check boxn
    print "<td align=\"center\" width=\"50%\"><b>Show Done?<b>\n";
    my $checked = '';
    if ($showdone == 1) {
        $checked = 'checked';
    }
    print "<input type=\"checkbox\" name=\"showdone\" value=\"showdone\"";
    print " $checked /> </td>\n";

    # print submit button and start of add items table
    print <<EOF;
    </tr></table>
    <p><table align=\"center\">
    <tr><td align="center" valign="center">
    <input type="submit" value="   Update   " name="update"></td></tr>
    </table><hr />
    <table align="center">
    <tr><th>Priority</th><th>Date</th><th align="left">Description</th>
EOF
    # print priority menu;
    print "<tr><td><select name=\"newprior\">\n";
    my $i;
    foreach $i (1..5) {                # priorities 1 to 5
        if ($newerror and param('newprior') == $i) {
            $checked = 'selected';
        }
        print "<option $checked>$i</option>\n";
    }
    print "</select></td>\n";

    # print date and description cells; may be different in case of
    # errors
    my $newdate = '';
    my $newdesc = '';
    print "<td align=\"center\"><input type=\"text\" name=\"newdate\"";
    if ($newerror) {                # has an error occurred?
        $newdate = "***" . param('newdate');
        $newdesc = param('newdesc');
    } 
    print "value=\"$newdate\" size=\"10\"></td> \n";
    # description cell; preserve old value if error
    print "<td><input type=\"text\" name=\"newdesc\" value=\"$newdesc\"";
    print "size=\"50\"></td></tr></table><table align=\"center\">\n";

    # and finish up
    print <<EOF;
    <tr><td align="center" valign="center">
    <input type="submit" value="Add New Item" name="additems" /></td></tr>
    </table></form>
EOF
}

# display each line of the data.  Data is already sorted; this just
# prints an inidividual record
sub display_data {
    my %rec = @_;                # record to print 

    # don't show done items if Show Done is unchecked
    # BUT include their settings anyhow (otherwise its too
    # difficult to figure out what's shown and changed versus
    # what's hidden
    if ($showdone == 0 and $rec{'done'}) {
        print "<input type=\"hidden\" name=\"done", $rec{'id'};
        print "\" />\n";
        next;
    }
    # make 1 priority items print in red
    my $bgcolor = '';               # priority items are red, all others ''
    if ($rec{'prior'} == 1) {
        $bgcolor = "bgcolor=\"red\"";
    }

    # Is it done or not?
    my $checked = '';                # done items are checked
    if ($rec{'done'}) {
        $checked = 'checked="checked"';
    }

    print "<!-- ID: ", $rec{id}, " -->\n";

    print "<tr>\n";                # start row

    # done boxes
    print "<td width=\"10%\" align=\"center\" $bgcolor>";
    print "<input type=\"checkbox\" name=\"done", $rec{'id'};
    print "\" $checked /></td>\n"; 

    # priority menus
    print "<td width=\"10%\" align=\"center\" $bgcolor>";
    print "<select name=\"prior", $rec{'id'}, "\">\n";
    my $select = '';
    my $i;
    foreach $i (1..5) {                # priorities 1 to 5
        $checked = '';
        if ($rec{'prior'} == $i) {
            $checked = 'selected';
        }
        print "<option value=\"$i\" $checked>$i</option>\n";
        $select = '';
    }
    print "</select></td>\n";

    # dates
    print "<td $bgcolor width=\"10%\" align=\"checked\">";
    print "<input type=\"text\" size=\"10\" name=\"date", $rec{'id'}, "\" ";
    print "value=\"", $rec{'date'}, "\" /></td>\n";

    # descriptions
    print "<td $bgcolor><input type=\"text\" name=\"desc", $rec{'id'};
    print "\" size=\"50\" value=\"", $rec{'desc'}, "\" /></td>\n";

    # Remove boxes
    print "<td $bgcolor align=\"center\">";
    print "<input type=\"checkbox\" name=\"r", $rec{'id'}, "\" /></td>";

    # end row
    print "</tr>\n\n";
}

# update all values in case changes were made
sub update_data {
    my $error = 0;                # error checking
    # check to see if showdone is selected;
    if (defined param('showdone')) {
        $showdone = 1;         
    } else {
        $showdone = 0;
    }   
    
    # get currrent sortby value
    $sortby = param('sortby');

    foreach (@data) {
        my $id = $_->{'id'};        # not the global $id

        # Entries that are marked done cannot be changed (usability
        # assumption).  So if an entry is marked done, and hasn't
        # been changed to not-done, we can skip checking any of the
        # rest of its data.
        if ($_->{'done'} == 1 && defined param('done' . $id)) {
            next;
        }

        # All newly done items.
        if (defined param('done' . $id)) {
            $_->{'done'} = 1;
        } else {
            $_->{'done'} = 0;
        }
        # dates.  check for weird date
        if (param('date' . $id) ne $_->{'date'}) {
            $error = check_date(param('date' . $id));
            if ($error) {
                $_->{'date'} = "*** " . param('date' . $id);
            } else {
                $_->{'date'} = param('date' . $id);
            }
        }

        # priorities, descriptions, change only if different
        my $thing;
        foreach $thing ('prior', 'desc') {
            if (param($thing . $id) ne $_->{$thing}) {
                $_->{$thing} = param($thing . $id);
            }
        }
    }
    return $error;
}

# remove items by redoing the @data list
sub remove_selected {
    my @newdata = ();
    foreach (@data) {
        my $id = $_->{'id'};        # also not the global id

        if (!defined param('r' . $id)) {
            push @newdata, $_;        # $_ is the reference
        } 
    }
    @data = @newdata;                # get rid of removed items
}

# add a new item.  This is only called if check_date has already said OK
sub add_item {
    my %newrec = ();

    $newrec{'desc'} = param('newdesc');
    $newrec{'date'} = param('newdate');
    $newrec{'prior'} = param('newprior');
    $newrec{'done'} = 0;

    my $max_id = 0;

    foreach my $datum (@data) {
        if ($datum->{'id'} > $max_id) {
            $max_id = $datum->{'id'};
        }
    }

    $newrec{'id'} = ++$max_id;        # global ID + 1
    push @data, { %newrec };
}

# dates must be in XX/XX/XX format
sub check_date {
    my $date = shift;
    # MM/DD/YYYY, MM and DD can be 0 or 1 char, but YYYY must be four
    # ending whitespace is OK.
    if ($date !~ /^(\d{1,2})\/(\d{1,2})\/(\d{4})\s*$/) {
        return 1;                # error!
    }
    return 1 if ($1 > 12);
    return 1 if ($2 > 31);

    return 0;                    # OK date
}

# rewrite data file
sub write_data {
    open(DATA, ">$listdata") or die "Can't open list data: $!.";
    foreach (@data) {
        my %rec = %$_;

        foreach ('id', 'desc', 'date','prior','done') { 
            print DATA "$_=$rec{$_}\n";
        }
        print DATA "---\n";
    }
    close(DATA);
}

# I use MM/DD/YY format for dates.  To sort by date you need to
# convert this format back into Perl's seconds-since-1900 format.
# the Time::Local module and the timelocal func do this.
sub date2time {
    my $date = shift;
    if ($date =~ /^\*\*\*/) {        # error formatting, sort to top
        return 0;
    } else {
        my ($m,$d,$y) = split(/\//,$date);
        $m--;                   # months start from 0 in perl's time format
        return timelocal(0,0,0,$d,$m,$y);
    }
}

